@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2018. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -38,7 +38,7 @@
 -export([appcall/4]).
 
 -import(lists, [reverse/1,flatten/1,sublist/3,sort/1,keysort/2,
-		max/1,min/1,foreach/2,foldl/3,flatmap/2]).
+		foreach/2,foldl/3,flatmap/2]).
 -import(io, [format/1, format/2]).
 
 %%-----------------------------------------------------------------------
@@ -840,12 +840,9 @@ bi(I) ->
 m(M) ->
     L = M:module_info(),
     {exports,E} = lists:keyfind(exports, 1, L),
-    Time = get_compile_time(L),
     COpts = get_compile_options(L),
     format("Module: ~w~n", [M]),
     print_md5(L),
-    format("Compiled: "),
-    print_time(Time),
     print_object_file(M),
     format("Compiler options:  ~p~n", [COpts]),
     format("Exports: ~n",[]), print_exports(keysort(1, E)).
@@ -864,12 +861,6 @@ print_md5(L) ->
         _ -> ok
     end.
 
-get_compile_time(L) ->
-    case get_compile_info(L, time) of
-	{ok,Val} -> Val;
-	error -> notime
-    end.
-
 get_compile_options(L) ->
     case get_compile_info(L, options) of
 	{ok,Val} -> Val;
@@ -910,25 +901,6 @@ split_print_exports([{F1, A1}|T1], [{F2, A2} | T2]) ->
     split_print_exports(T1, T2);
 split_print_exports([], []) -> ok.
 
-print_time({Year,Month,Day,Hour,Min,_Secs}) ->
-    format("~s ~w ~w, ", [month(Month),Day,Year]),
-    format("~.2.0w:~.2.0w~n", [Hour,Min]);
-print_time(notime) ->
-    format("No compile time info available~n",[]).
-
-month(1) -> "January";
-month(2) -> "February";
-month(3) -> "March";
-month(4) -> "April";
-month(5) -> "May";
-month(6) -> "June";
-month(7) -> "July";
-month(8) -> "August";
-month(9) -> "September";
-month(10) -> "October";
-month(11) -> "November";
-month(12) -> "December".
-
 %% Just because we can't eval receive statements...
 -spec flush() -> 'ok'.
 
@@ -1056,11 +1028,15 @@ ls() ->
 -spec ls(Dir) -> 'ok' when
       Dir :: file:name().
 
-ls(Dir) ->
-    case file:list_dir(Dir) of
+ls(Dir0) ->
+    case file:list_dir(Dir0) of
 	{ok, Entries} ->
 	    ls_print(sort(Entries));
 	{error, enotdir} ->
+            Dir = if
+                      is_list(Dir0) -> lists:flatten(Dir0);
+                      true -> Dir0
+                  end,
 	    ls_print([Dir]);
 	{error, Error} ->
 	    format("~ts\n", [file:format_error(Error)])
@@ -1068,7 +1044,7 @@ ls(Dir) ->
 
 ls_print([]) -> ok;
 ls_print(L) ->
-    Width = min([max(lengths(L, [])), 40]) + 5,
+    Width = erlang:min(max_length(L, 0), 40) + 5,
     ls_print(L, Width, 0).
 
 ls_print(X, Width, Len) when Width + Len >= 80 ->
@@ -1080,8 +1056,12 @@ ls_print([H|T], Width, Len) ->
 ls_print([], _, _) ->
     io:nl().
 
-lengths([H|T], L) -> lengths(T, [length(H)|L]);
-lengths([], L)    -> L.
+max_length([H|T], L) when is_atom(H) ->
+    max_length([atom_to_list(H)|T], L);
+max_length([H|T], L) ->
+    max_length(T, erlang:max(length(H), L));
+max_length([], L) ->
+    L.
 
 w(X) ->
     io_lib:write(X).
